from enum import Enum


class CardNetwork(str, Enum):
    atm = 'atm'
    visa = 'visa'


class CardStatus(str, Enum):
    active = 'active'
    blocked = 'blocked'
    created = 'created'
    deactivated = 'deactivated'
    printing = 'printing'


class CardTransactionType(str, Enum):
    auth = 'auth'
    capture = 'capture'
    expiration = 'expiration'
    refund = 'refund'
    void = 'void'
    chargeback = 'chargeback'
    fast_funds = 'fast_funds'
    push = 'push'
    push_confirmation = 'push_confirmation'


class CardFraudType(str, Enum):
    authorize = 'authorize'
    rejected = 'rejected'
    uncertain = 'uncertain'


class CardType(str, Enum):
    physical = 'physical'
    virtual = 'virtual'


class CardIssuer(str, Enum):
    accendo = 'accendo'
    cuenca = 'cuenca'


class IssuerNetwork(str, Enum):
    mastercard = 'Mastercard'
    visa = 'Visa'


class CardFundingType(str, Enum):
    credit = 'credit'
    debit = 'debit'


class DepositNetwork(str, Enum):
    cash = 'cash'
    internal = 'internal'
    spei = 'spei'


class ServiceProviderCategory(str, Enum):
    cable = 'cable'
    credit_card = 'credit_card'
    electricity = 'electricity'
    gas = 'gas'
    internet = 'internet'
    landline_telephone = 'landline_telephone'
    mobile_telephone_postpaid = 'mobile_telephone_postpaid'
    mobile_telephone_prepaid = 'mobile_telephone_prepaid'
    satelite_television = 'satelite_television'
    water = 'water'


class TransactionStatus(str, Enum):
    created = 'created'
    submitted = 'submitted'
    in_review = 'in_review'
    succeeded = 'succeeded'
    failed = 'failed'


class TransferNetwork(str, Enum):
    internal = 'internal'
    spei = 'spei'


class CommissionType(str, Enum):
    card_request = 'card_request'
    cash_deposit = 'cash_deposit'
    outgoing_spei = 'outgoing_spei'
    card_shipping = 'card_shipping'
    arteria = 'arteria'
    connect = 'connect'


class EntryType(str, Enum):
    credit = 'credit'
    debit = 'debit'


class CardErrorType(str, Enum):
    blocked = 'blocked'
    insufficient_founds = 'insufficient_founds'
    notification = 'notification'
    notification_deactivated_card = 'notification_deactivated_card'
    contactless_amount_limit = 'contactless_amount_limit'
    communication = 'communication'
    fraud_detection = 'fraud_detection'
    fraud_detection_uncertain = 'fraud_detection_uncertain'
    invalid_pin = 'invalid_pin'
    cash_advance_daily_limit = 'cash_advance_daily_limit'


class FileFormat(str, Enum):
    pdf = 'application/pdf'
    xml = 'application/xml'
    any = '*/*'


class TrackDataMethod(str, Enum):
    not_set = 'not-set'
    terminal = 'terminal'
    manual = 'manual'
    unknown = 'unknown'
    contactless = 'contactless'
    fall_back = 'fall_back'
    magnetic_stripe = 'magnetic_stripe'
    recurring_charge = 'recurring_charge'


class PosCapability(str, Enum):
    not_set = 'not-set'
    unknown = 'unknown'
    pin_accepted = 'pin_accepted'
    pin_not_accepted = 'pin_not_accepted'
    pin_pad_down = 'pin_pad_down'
    reserved = 'reserved'


class AuthorizerTransaction(str, Enum):
    advice = 'advice'
    normal_purchase = 'normal_purchase'
    cash_advance = 'cash_advance'
    returns = 'returns'
    balance_inquiry = 'balance_inquiry'
    purchase_with_cashback = 'purchase_with_cashback'
    not_defined = 'not_defined'
    mail_or_phone_order = 'mail_or_phone_order'
    change_pin = 'change_pin'
    notification = 'notification'
    card_validation = 'card_validation'
    check_out = 'check_out'
    re_authorization = 're_authorization'
    re_returns = 're_returns'
    fast_funds = 'fast_funds'
    fast_funds_reverse = 'fast_funds_reverse'


class UserCardNotification(str, Enum):
    balance_inquiry = 'balance_inquiry'
    card_blocked = 'card_blocked'
    monthly_purchases = 'monthly_purchases'


class CardDesign(str, Enum):
    classic = 'classic'
    travesia = 'travesia'
    limited_edition = 'limited_edition'


class CardPackaging(str, Enum):
    manual_local = 'NL'
    manual_nonlocal = 'NF'
    automated_local = 'RL'
    automated_nonlocal = 'RF'
    automated_batch_shipping = 'RE'


class CardholderVerificationMethod(str, Enum):
    # Describes how the cardholder verified their
    # identity (PIN, signature, with app, etc.).
    unknown = 'unknown'
    signature = 'signature'
    pin = 'pin'
    not_attended = 'not_attended'
    mail_or_phone = 'mail_or_phone'
    quick_payment_service = 'quick_payment_service'
    contactless = 'contactless'
    app_confirmation = 'app_confirmation'


class WalletTransactionType(str, Enum):
    deposit = 'deposit'
    withdrawal = 'withdrawal'


class SavingCategory(str, Enum):
    general = 'general'
    home = 'home'
    vehicle = 'vehicle'
    travel = 'travel'
    clothing = 'clothing'
    other = 'other'
    medical = 'medical'
    accident = 'accident'
    education = 'education'


class EcommerceIndicator(str, Enum):
    """
    Adquirente graba
    Emisor lee
    0, “ “ = No es una transacción de comercio electrónico
    1 = Transacción MOTO
    2 = MOTO Indicator – Recurring Transaction
    3 = MOTO Indicator – Installment Payment
    4 = MOTO Indicator – Deferred Transaction
    5 = Comercio seguro, titular autenticado (3D Secure)
    6 = Comercio seguro, titular no autenticado (3D Secure)
    7 = Autenticación 3D Secure no realizada
    8 – La transacción no incluye el uso de ningún cifrado de transacción
        como SSL, no se realiza autenticación.
    """

    not_ecommerce = '0'
    moto_transaction = '1'
    moto_recurring_transaction = '2'
    moto_installment_payment = '3'
    moto_deferred_transaction = '4'
    authenticated_3ds = '5'
    not_authenticated_3ds = '6'
    not_3ds = '7'
    not_secure_transaction = '8'


class TransactionTokenValidationStatus(str, Enum):
    pending = 'pending'
    accepted = 'accepted'
    rejected = 'rejected'


class KYCFileType(str, Enum):
    dni = 'dni'
    ine = 'ine'
    passport = 'passport'
    residency = 'residency'
    matricula_consular = 'matricula_consular'
    proof_of_liveness = 'proof_of_liveness'
    proof_of_address = 'proof_of_address'
    rfc = 'rfc'


class Gender(str, Enum):
    female = 'female'
    male = 'male'


class State(str, Enum):
    NE = 'NE'  # Nacido en el Extranjero
    AS = 'AS'  # Aguascalientes
    BC = 'BC'  # Baja California
    BS = 'BS'  # Baja California Sur
    CC = 'CC'  # Campeche
    CS = 'CS'  # Chiapas
    CH = 'CH'  # Chihuahua
    CL = 'CL'  # Coahuila
    CM = 'CM'  # Colima
    DF = 'DF'  # Ciudad de México
    DG = 'DG'  # Durango
    GT = 'GT'  # Guanajuato
    GR = 'GR'  # Guerrero
    HG = 'HG'  # Hidalgo
    JC = 'JC'  # Jalisco
    MC = 'MC'  # México
    MN = 'MN'  # Michoacan
    MS = 'MS'  # Morelos
    NT = 'NT'  # Nayarit
    NL = 'NL'  # Nuevo León
    OC = 'OC'  # Oaxaca
    PL = 'PL'  # Puebla
    QT = 'QT'  # Querétaro
    QR = 'QR'  # Quintana Roo
    SP = 'SP'  # San Luis Potosí
    SL = 'SL'  # Sinaloa
    SR = 'SR'  # Sonora
    TC = 'TC'  # Tabasco
    TL = 'TL'  # Tlaxcala
    TS = 'TS'  # Tamaulipas
    VZ = 'VZ'  # Veracruz
    YN = 'YN'  # Yucatán
    ZS = 'ZS'  # Zacatecas


# Country enum was built with pycountry==22.1.10
# ISO 3166-1  Alpha 2
# https://github.com/flyingcircusio/pycountry/blob/af4e4a8b14388e50d5d796e7a732efd129032fd8/src/pycountry/databases/iso3166-1.json
class Country(str, Enum):
    AW = 'AW'  # Aruba
    AF = 'AF'  # Afghanistan
    AO = 'AO'  # Angola
    AI = 'AI'  # Anguilla
    AX = 'AX'  # Åland Islands
    AL = 'AL'  # Albania
    AD = 'AD'  # Andorra
    AE = 'AE'  # United Arab Emirates
    AR = 'AR'  # Argentina
    AM = 'AM'  # Armenia
    AS = 'AS'  # American Samoa
    AQ = 'AQ'  # Antarctica
    TF = 'TF'  # French Southern Territories
    AG = 'AG'  # Antigua and Barbuda
    AU = 'AU'  # Australia
    AT = 'AT'  # Austria
    AZ = 'AZ'  # Azerbaijan
    BI = 'BI'  # Burundi
    BE = 'BE'  # Belgium
    BJ = 'BJ'  # Benin
    BQ = 'BQ'  # Bonaire, Sint Eustatius and Saba
    BF = 'BF'  # Burkina Faso
    BD = 'BD'  # Bangladesh
    BG = 'BG'  # Bulgaria
    BH = 'BH'  # Bahrain
    BS = 'BS'  # Bahamas
    BA = 'BA'  # Bosnia and Herzegovina
    BL = 'BL'  # Saint Barthélemy
    BY = 'BY'  # Belarus
    BZ = 'BZ'  # Belize
    BM = 'BM'  # Bermuda
    BO = 'BO'  # Bolivia, Plurinational State of
    BR = 'BR'  # Brazil
    BB = 'BB'  # Barbados
    BN = 'BN'  # Brunei Darussalam
    BT = 'BT'  # Bhutan
    BV = 'BV'  # Bouvet Island
    BW = 'BW'  # Botswana
    CF = 'CF'  # Central African Republic
    CA = 'CA'  # Canada
    CC = 'CC'  # Cocos (Keeling) Islands
    CH = 'CH'  # Switzerland
    CL = 'CL'  # Chile
    CN = 'CN'  # China
    CI = 'CI'  # Côte d'Ivoire
    CM = 'CM'  # Cameroon
    CD = 'CD'  # Congo, The Democratic Republic of the
    CG = 'CG'  # Congo
    CK = 'CK'  # Cook Islands
    CO = 'CO'  # Colombia
    KM = 'KM'  # Comoros
    CV = 'CV'  # Cabo Verde
    CR = 'CR'  # Costa Rica
    CU = 'CU'  # Cuba
    CW = 'CW'  # Curaçao
    CX = 'CX'  # Christmas Island
    KY = 'KY'  # Cayman Islands
    CY = 'CY'  # Cyprus
    CZ = 'CZ'  # Czechia
    DE = 'DE'  # Germany
    DJ = 'DJ'  # Djibouti
    DM = 'DM'  # Dominica
    DK = 'DK'  # Denmark
    DO = 'DO'  # Dominican Republic
    DZ = 'DZ'  # Algeria
    EC = 'EC'  # Ecuador
    EG = 'EG'  # Egypt
    ER = 'ER'  # Eritrea
    EH = 'EH'  # Western Sahara
    ES = 'ES'  # Spain
    EE = 'EE'  # Estonia
    ET = 'ET'  # Ethiopia
    FI = 'FI'  # Finland
    FJ = 'FJ'  # Fiji
    FK = 'FK'  # Falkland Islands (Malvinas)
    FR = 'FR'  # France
    FO = 'FO'  # Faroe Islands
    FM = 'FM'  # Micronesia, Federated States of
    GA = 'GA'  # Gabon
    GB = 'GB'  # United Kingdom
    GE = 'GE'  # Georgia
    GG = 'GG'  # Guernsey
    GH = 'GH'  # Ghana
    GI = 'GI'  # Gibraltar
    GN = 'GN'  # Guinea
    GP = 'GP'  # Guadeloupe
    GM = 'GM'  # Gambia
    GW = 'GW'  # Guinea-Bissau
    GQ = 'GQ'  # Equatorial Guinea
    GR = 'GR'  # Greece
    GD = 'GD'  # Grenada
    GL = 'GL'  # Greenland
    GT = 'GT'  # Guatemala
    GF = 'GF'  # French Guiana
    GU = 'GU'  # Guam
    GY = 'GY'  # Guyana
    HK = 'HK'  # Hong Kong
    HM = 'HM'  # Heard Island and McDonald Islands
    HN = 'HN'  # Honduras
    HR = 'HR'  # Croatia
    HT = 'HT'  # Haiti
    HU = 'HU'  # Hungary
    ID = 'ID'  # Indonesia
    IM = 'IM'  # Isle of Man
    IN = 'IN'  # India
    IO = 'IO'  # British Indian Ocean Territory
    IE = 'IE'  # Ireland
    IR = 'IR'  # Iran, Islamic Republic of
    IQ = 'IQ'  # Iraq
    IS = 'IS'  # Iceland
    IL = 'IL'  # Israel
    IT = 'IT'  # Italy
    JM = 'JM'  # Jamaica
    JE = 'JE'  # Jersey
    JO = 'JO'  # Jordan
    JP = 'JP'  # Japan
    KZ = 'KZ'  # Kazakhstan
    KE = 'KE'  # Kenya
    KG = 'KG'  # Kyrgyzstan
    KH = 'KH'  # Cambodia
    KI = 'KI'  # Kiribati
    KN = 'KN'  # Saint Kitts and Nevis
    KR = 'KR'  # Korea, Republic of
    KW = 'KW'  # Kuwait
    LA = 'LA'  # Lao People's Democratic Republic
    LB = 'LB'  # Lebanon
    LR = 'LR'  # Liberia
    LY = 'LY'  # Libya
    LC = 'LC'  # Saint Lucia
    LI = 'LI'  # Liechtenstein
    LK = 'LK'  # Sri Lanka
    LS = 'LS'  # Lesotho
    LT = 'LT'  # Lithuania
    LU = 'LU'  # Luxembourg
    LV = 'LV'  # Latvia
    MO = 'MO'  # Macao
    MF = 'MF'  # Saint Martin (French part)
    MA = 'MA'  # Morocco
    MC = 'MC'  # Monaco
    MD = 'MD'  # Moldova, Republic of
    MG = 'MG'  # Madagascar
    MV = 'MV'  # Maldives
    MX = 'MX'  # Mexico
    MH = 'MH'  # Marshall Islands
    MK = 'MK'  # North Macedonia
    ML = 'ML'  # Mali
    MT = 'MT'  # Malta
    MM = 'MM'  # Myanmar
    ME = 'ME'  # Montenegro
    MN = 'MN'  # Mongolia
    MP = 'MP'  # Northern Mariana Islands
    MZ = 'MZ'  # Mozambique
    MR = 'MR'  # Mauritania
    MS = 'MS'  # Montserrat
    MQ = 'MQ'  # Martinique
    MU = 'MU'  # Mauritius
    MW = 'MW'  # Malawi
    MY = 'MY'  # Malaysia
    YT = 'YT'  # Mayotte
    NA = 'NA'  # Namibia
    NC = 'NC'  # New Caledonia
    NE = 'NE'  # Niger
    NF = 'NF'  # Norfolk Island
    NG = 'NG'  # Nigeria
    NI = 'NI'  # Nicaragua
    NU = 'NU'  # Niue
    NL = 'NL'  # Netherlands
    NO = 'NO'  # Norway
    NP = 'NP'  # Nepal
    NR = 'NR'  # Nauru
    NZ = 'NZ'  # New Zealand
    OM = 'OM'  # Oman
    PK = 'PK'  # Pakistan
    PA = 'PA'  # Panama
    PN = 'PN'  # Pitcairn
    PE = 'PE'  # Peru
    PH = 'PH'  # Philippines
    PW = 'PW'  # Palau
    PG = 'PG'  # Papua New Guinea
    PL = 'PL'  # Poland
    PR = 'PR'  # Puerto Rico
    KP = 'KP'  # Korea, Democratic People's Republic of
    PT = 'PT'  # Portugal
    PY = 'PY'  # Paraguay
    PS = 'PS'  # Palestine, State of
    PF = 'PF'  # French Polynesia
    QA = 'QA'  # Qatar
    RE = 'RE'  # Réunion
    RO = 'RO'  # Romania
    RU = 'RU'  # Russian Federation
    RW = 'RW'  # Rwanda
    SA = 'SA'  # Saudi Arabia
    SD = 'SD'  # Sudan
    SN = 'SN'  # Senegal
    SG = 'SG'  # Singapore
    GS = 'GS'  # South Georgia and the South Sandwich Islands
    SH = 'SH'  # Saint Helena, Ascension and Tristan da Cunha
    SJ = 'SJ'  # Svalbard and Jan Mayen
    SB = 'SB'  # Solomon Islands
    SL = 'SL'  # Sierra Leone
    SV = 'SV'  # El Salvador
    SM = 'SM'  # San Marino
    SO = 'SO'  # Somalia
    PM = 'PM'  # Saint Pierre and Miquelon
    RS = 'RS'  # Serbia
    SS = 'SS'  # South Sudan
    ST = 'ST'  # Sao Tome and Principe
    SR = 'SR'  # Suriname
    SK = 'SK'  # Slovakia
    SI = 'SI'  # Slovenia
    SE = 'SE'  # Sweden
    SZ = 'SZ'  # Eswatini
    SX = 'SX'  # Sint Maarten (Dutch part)
    SC = 'SC'  # Seychelles
    SY = 'SY'  # Syrian Arab Republic
    TC = 'TC'  # Turks and Caicos Islands
    TD = 'TD'  # Chad
    TG = 'TG'  # Togo
    TH = 'TH'  # Thailand
    TJ = 'TJ'  # Tajikistan
    TK = 'TK'  # Tokelau
    TM = 'TM'  # Turkmenistan
    TL = 'TL'  # Timor-Leste
    TO = 'TO'  # Tonga
    TT = 'TT'  # Trinidad and Tobago
    TN = 'TN'  # Tunisia
    TR = 'TR'  # Turkey
    TV = 'TV'  # Tuvalu
    TW = 'TW'  # Taiwan, Province of China
    TZ = 'TZ'  # Tanzania, United Republic of
    UG = 'UG'  # Uganda
    UA = 'UA'  # Ukraine
    UM = 'UM'  # United States Minor Outlying Islands
    UY = 'UY'  # Uruguay
    US = 'US'  # United States
    UZ = 'UZ'  # Uzbekistan
    VA = 'VA'  # Holy See (Vatican City State)
    VC = 'VC'  # Saint Vincent and the Grenadines
    VE = 'VE'  # Venezuela, Bolivarian Republic of
    VG = 'VG'  # Virgin Islands, British
    VI = 'VI'  # Virgin Islands, U.S.
    VN = 'VN'  # Viet Nam
    VU = 'VU'  # Vanuatu
    WF = 'WF'  # Wallis and Futuna
    WS = 'WS'  # Samoa
    YE = 'YE'  # Yemen
    ZA = 'ZA'  # South Africa
    ZM = 'ZM'  # Zambia
    ZW = 'ZW'  # Zimbabwe


class VerificationStatus(str, Enum):
    not_verified = 'not_verified'
    created = 'created'
    submitted = 'submitted'
    rejected = 'rejected'
    succeeded = 'succeeded'
    review_needed = 'review_needed'
    upload_again = 'upload_again'


class EventType(str, Enum):
    created = 'created'
    succeeded = 'succeeded'
    failed = 'failed'
    updated = 'updated'
    deactivated = 'deactivated'


class UserStatus(str, Enum):
    active = 'active'
    deactivated = 'deactivated'
    fraud = 'fraud'
    in_review = 'in_review'
    pld_blocked = 'pld_blocked'


class SessionType(str, Enum):
    registration = 'session.registration'
    show_card = 'session.show_card'
    account_entries = 'session.account_entries'
    download_file = 'session.download_file'
    upload_file = 'session.upload_file'


class WebhookObject(str, Enum):
    user = 'user'
    transaction = 'transaction'


class WebhookEventType(str, Enum):
    create = 'create'
    update = 'update'
    delete = 'delete'


class WebhookEvent(str, Enum):
    card_transaction_create = 'card_transaction.create'
    card_transaction_update = 'card_transaction.update'
    user_create = 'user.create'
    user_update = 'user.update'
    user_delete = 'user.delete'
    transaction_create = 'transaction.create'
    transaction_update = 'transaction.update'
    cash_deposit_create = 'cash_deposit.create'
    cash_deposit_update = 'cash_deposit.update'
    bank_account_create = 'bank_account.create'
    bank_account_update = 'bank_account.update'


class VerificationType(str, Enum):
    phone = 'phone'
    email = 'email'


class Language(str, Enum):
    en = 'en'
    es = 'es'


class PlatformType(str, Enum):
    bridge = 'bridge'
    connect = 'connect'
    spei = 'spei'


class TermsOfService(str, Enum):
    arteria = 'arteria'
    ifpe = 'ifpe'
    tarjetas_cuenca = 'tarjetas_cuenca'
    portal = 'portal'


class FileExtension(str, Enum):
    mp4 = 'mp4'
    mov = 'mov'
    v_3gpp = '3gpp'  # 3GPP
    jpg = 'jpg'
    jpeg = 'jpeg'
    png = 'png'
    pdf = 'pdf'


class BankAccountStatus(str, Enum):
    created = 'created'
    succeeded = 'succeeded'
    failed = 'failed'



import pytest
from pydantic import BaseModel, ValidationError

from cuenca_validations.errors import CardBinValidationError
from cuenca_validations.types import PaymentCardNumber, StrictPaymentCardNumber

VALID_BBVA = '4772130000000003'
INVALID_BIN = '4050000000000001'


def test_valid_bin():
    class PaymentCard(BaseModel):
        card_number: PaymentCardNumber

    pc = PaymentCard(card_number=VALID_BBVA)
    assert pc.card_number.bin == '477213'
    assert pc.card_number.bank_code == '40012'


def test_invalid_bin():
    class PaymentCard(BaseModel):
        card_number: StrictPaymentCardNumber

    with pytest.raises(ValidationError) as exc_info:
        PaymentCard(card_number=INVALID_BIN)
    assert exc_info.value.errors()[0] == dict(
        loc=('card_number',),
        type=f'value_error.{CardBinValidationError.code}',
        msg=CardBinValidationError.msg_template,
    )


def test_valid_bin_strict_payment():
    cn = StrictPaymentCardNumber.validate_bin(PaymentCardNumber(VALID_BBVA))
    assert cn == VALID_BBVA



import datetime as dt
import json
from dataclasses import dataclass
from enum import Enum

import pytest
from freezegun import freeze_time
from pydantic import BaseModel, ValidationError

from cuenca_validations.types import (
    Address,
    CardQuery,
    JSONEncoder,
    QueryParams,
    SantizedDict,
    SessionRequest,
    TransactionStatus,
    digits,
    get_state_name,
)
from cuenca_validations.types.enums import (
    Country,
    EcommerceIndicator,
    SessionType,
    State,
)
from cuenca_validations.types.requests import (
    ApiKeyUpdateRequest,
    BankAccountValidationRequest,
    ChargeRequest,
    CurpValidationRequest,
    EndpointRequest,
    EndpointUpdateRequest,
    IdentityUpdateRequest,
    LimitedWalletRequest,
    SavingRequest,
    SavingUpdateRequest,
    UserCardNotificationRequest,
    UserCredentialUpdateRequest,
    UserRequest,
    UserUpdateRequest,
    VerificationAttemptRequest,
    VerificationRequest,
)

today = dt.date.today()
now = dt.datetime.now()
utcnow = now.astimezone(dt.timezone.utc)


class TestEnum(Enum):
    zero = 0


@dataclass
class TestClass:
    uno: str

    def to_dict(self):
        return dict(uno=self.uno, dos='dos')


def test_dict():
    model = QueryParams(count=1, created_before=now)
    assert model.dict() == dict(count=1, created_before=utcnow.isoformat())


def test_dict_with_exclude():
    model = QueryParams(count=1, created_before=now, user_id='USXXXX')
    assert model.dict(exclude={'user_id'}) == dict(
        count=1, created_before=utcnow.isoformat()
    )


def test_dict_with_exclude_unset():
    model = QueryParams(count=1, created_before=now)
    assert model.dict(exclude_unset=False) == dict(
        count=1, created_before=utcnow.isoformat(), page_size=100
    )


def test_sanitized_dict():
    assert SantizedDict(
        status=TransactionStatus.succeeded,
        time=now,
        hello='there',
        dates=[now],
    ) == dict(
        status='succeeded',
        time=utcnow.isoformat(),
        hello='there',
        dates=[utcnow.isoformat()],
    )


@pytest.mark.parametrize(
    'count, truth',
    [
        (1, True),
        ('1', True),
        (True, True),
        (False, False),
        (0, False),
        ('0', False),
    ],
)
def test_count(count, truth):
    q = QueryParams(count=count)
    assert q.count is truth


@pytest.mark.parametrize(
    'value, result',
    [
        (TestEnum.zero, 0),
        (today, today.isoformat()),
        (now, utcnow.isoformat()),
        (TestClass(uno='uno'), dict(uno='uno', dos='dos')),
    ],
)
def test_json_encoder(value, result):
    to_encode = dict(value=value)
    encoded = json.dumps(to_encode, cls=JSONEncoder)
    decoded = json.loads(encoded)

    assert decoded['value'] == result


def test_invalid_class():
    """
    For a class that doesn't have a `to_dict` method and it is not a type of
    `date` nor `Enum`, will use the default `json.JSONEncoder` method which
    raises a `TypeError`.
    """

    class ClassWithoutToDict:
        ...

    invalid_class = ClassWithoutToDict()
    with pytest.raises(TypeError):
        json.dumps(invalid_class, cls=JSONEncoder)


class Accounts(BaseModel):
    number: digits(5, 8)  # type: ignore


def test_only_digits():
    acc = Accounts(number='123456')
    assert acc.number == '123456'


@pytest.mark.parametrize(
    'number, error',
    [
        ('123', 'value_error.any_str.min_length'),
        ('1234567890', 'value_error.any_str.max_length'),
        ('no_123', 'value_error.digits'),
    ],
)
def test_invalid_digits(number, error):
    with pytest.raises(ValueError) as exception:
        Accounts(number=number)
    assert error in str(exception)


def test_card_query_exp_cvv_if_number_set():
    values = dict(number='123456', exp_month=1, exp_year=2026)
    card_query = CardQuery(**values)
    assert all(
        getattr(card_query, key) == value for key, value in values.items()
    )


@pytest.mark.parametrize(
    'input_value',
    [
        (dict(exp_month=1)),
        (dict(exp_year=2026)),
        (dict(cvv2='123')),
        (dict(cvv='123')),
    ],
)
def test_card_query_exp_cvv_if_number_not_set(input_value):
    with pytest.raises(ValueError):
        CardQuery(**input_value)


def test_exclude_none_in_dict():
    request = ApiKeyUpdateRequest(user_id='US123')
    assert request.dict() == dict(user_id='US123')


def test_update_one_property_at_a_time_request():
    with pytest.raises(ValueError):
        UserCredentialUpdateRequest(user_id='US123', password='123456')

    req = UserCredentialUpdateRequest(password='123456')
    assert not req.is_active and req.password == '123456'

    req = UserCredentialUpdateRequest(is_active=True)
    assert req.is_active and not req.password


@pytest.mark.parametrize(
    'data,expected_dict',
    [
        (dict(password='123456'), dict(password='123456', is_active=None)),
        (dict(is_active=True), dict(password=None, is_active=True)),
        (dict(), dict(password=None, is_active=None)),
    ],
)
def test_update_credential_update_request_dict(data, expected_dict):
    req = UserCredentialUpdateRequest(**data)
    assert req.dict() == expected_dict


def test_card_transaction_requests():
    data = dict(
        card_id='CA123',
        user_id='US123',
        amount=100,
        merchant_name='visa',
        merchant_data='0279288357            00012558',
        merchant_type='wtype',
        currency_code='458',
        prosa_transaction_id='12345',
        authorizer_number='123456',
        retrieval_reference='who ks',
        transaction_type='normal_purchase',
        card_type='virtual',
        card_status='active',
        track_data_method='terminal',
        pos_capability='pin_accepted',
        is_cvv=False,
        get_balance=False,
        issuer='Mastercard',
        ecommerce_indicator='0',
    )
    ChargeRequest(**data)

    # Validate atm_fee optional
    data['atm_fee'] = 1800
    request = ChargeRequest(**data)
    assert request.ecommerce_indicator is EcommerceIndicator.not_ecommerce
    # missing fields
    with pytest.raises(ValidationError):
        UserCardNotificationRequest(**data)

    # invalid fields
    data['atm_fee'] = -1
    with pytest.raises(ValidationError):
        ChargeRequest(**data)
    data['amount'] = -1
    with pytest.raises(ValidationError):
        UserCardNotificationRequest(**data)


def test_saving_request():
    dt_now = dt.datetime.now()
    data = dict(
        name='My car',
        category='vehicle',
        goal_amount=66600,
        goal_date=dt_now + dt.timedelta(days=1),
    )
    SavingRequest(**data)

    data['goal_amount'] = -1000
    with pytest.raises(ValidationError):
        SavingRequest(**data)

    data['goal_amount'] = 66600
    data['goal_date'] = dt_now
    with pytest.raises(ValidationError):
        SavingRequest(**data)


def test_saving_update_request():
    data = dict(
        name='Mt home ',
        category='home',
        goal_amount=1000,
    )
    SavingUpdateRequest(**data)
    data['goal_date'] = dt.datetime(2000, 1, 1)
    with pytest.raises(ValidationError):
        SavingUpdateRequest(**data)


def test_address_validation():
    data = dict(
        full_name='Varsovia 36, Col Cuahutemoc',
    )
    assert Address(**data)
    with pytest.raises(ValueError) as v:
        Address(**dict())
    assert 'required street' in str(v)
    data = dict(street='somestreet')
    with pytest.raises(ValueError) as v:
        Address(**data)
    assert 'required ext_number' in str(v)
    data = dict(
        street='varsovia',
        ext_number='36',
        state=State.DF,
        country=Country.MX,
    )
    assert Address(**data)


@freeze_time('2022-01-01')
def test_user_request():
    request = dict(
        id=None,
        curp='ABCD920604HDFSRN03',
        phone_number='+525555555555',
        email_address='email@email.com',
        profession='worker',
        status='active',
        address=dict(
            street='calle 1',
            ext_number='2',
            int_number='3',
            colonia='Juarez',
            postal_code='09900',
            state=State.DF.value,
            country=Country.MX,
            city='Obrera',
            full_name=None,
        ),
        phone_verification_id='VE12345678',
        email_verification_id='VE0987654321',
        required_level=3,
        terms_of_service=None,
    )
    assert UserRequest(**request).dict() == request

    # changing curp so user is underage
    request['curp'] = 'ABCD060604HDFSRN03'
    with pytest.raises(ValueError) as v:
        UserRequest(**request)
        assert 'User does not meet age requirement.' in str(v)


@freeze_time('2022-01-01')
def test_curp_validation_request():
    request = dict(
        names='Pedro',
        first_surname='Páramo',
        second_surname=None,
        date_of_birth=dt.date(1917, 5, 17),
        state_of_birth=State.DF.value,
        gender='male',
        manual_curp='ABCD920604HDFSRN03',
        country_of_birth='MX',
    )

    with pytest.raises(ValueError) as v:
        CurpValidationRequest()
    assert (
        'values required: names,first_surname,date_of_birth,'
        'country_of_birth,gender' in str(v)
    )

    req_curp = CurpValidationRequest(**request)
    assert req_curp.dict() == request

    request['date_of_birth'] = dt.date(2006, 5, 17)

    with pytest.raises(ValueError) as v:
        CurpValidationRequest(**request)
    assert 'User does not meet age requirement.' in str(v)

    # changing date of birth so user is underage
    request['date_of_birth'] = dt.date(1917, 5, 17)
    del request['state_of_birth']

    with pytest.raises(ValueError) as v:
        CurpValidationRequest(**request)
    assert 'state_of_birth required' in str(v)


def test_user_update_request():
    request = dict(
        beneficiaries=[
            dict(
                name='Pedro Pérez',
                birth_date=dt.date(2020, 1, 1),
                phone_number='+525555555555',
                user_relationship='brother',
                percentage=50,
            ),
            dict(
                name='José Pérez',
                birth_date=dt.date(2020, 1, 2),
                phone_number='+525544444444',
                user_relationship='brother',
                percentage=50,
            ),
        ]
    )
    update_req = UserUpdateRequest(**request)
    beneficiaries = [b.dict() for b in update_req.beneficiaries]
    assert beneficiaries == request['beneficiaries']

    request['beneficiaries'] = [
        dict(
            name='Pedro Pérez',
            birth_date=dt.date(2020, 1, 1).isoformat(),
            phone_number='+525555555555',
            user_relationship='brother',
            percentage=50,
        ),
    ]
    assert UserUpdateRequest(**request)

    request['beneficiaries'] = [
        dict(
            name='Pedro Pérez',
            birth_date=dt.date(2020, 1, 1).isoformat(),
            phone_number='+525555555555',
            user_relationship='brother',
            percentage=101,
        ),
    ]
    with pytest.raises(ValueError) as v:
        UserUpdateRequest(**request)
        assert 'The total percentage is more than 100.' in str(v)

    tos_request = dict(
        terms_of_service=dict(
            version='2022-01-01',
            ip='127.0.0.1',
            location='1111,1111',
            type='ifpe',
        )
    )
    UserUpdateRequest(**tos_request)

    tos_request = dict(
        terms_of_service=dict(
            version='2022-01-01',
            ip='2001:0db8:0000:0000:0000:ff00:0042:8329',
            location='1111,1111',
            type='ifpe',
        )
    )
    UserUpdateRequest(**tos_request)

    kyc_request = dict(
        govt_id=dict(
            type='ine',
            uri_front='/files/FI-123',
            uri_back='/files/FI-456',
        )
    )
    UserUpdateRequest(**kyc_request)

    # chagning to invalid request
    tos_request['terms_of_service']['ip'] = 'not valid ip'
    with pytest.raises(ValueError) as v:
        UserUpdateRequest(**tos_request)
    assert 'not valid ip' in str(v.value)


def test_session_request():
    data = dict(
        user_id='sadsa', type=SessionType.registration, success_url='no url'
    )
    with pytest.raises(ValidationError):
        SessionRequest(**data)
    data['success_url'] = 'http://url.com'
    assert SessionRequest(**data)


def test_endpoint_request():
    data = dict(url='bad url', events=['user.create'])
    with pytest.raises(ValidationError):
        EndpointRequest(**data)
    data['url'] = 'http://url.com'
    assert EndpointRequest(**data)


def test_endpoint_update_request():
    data = dict(is_enable=True, events=['user.read'])
    with pytest.raises(ValidationError):
        EndpointUpdateRequest(**data)
    data['events'] = ['user.update']
    assert EndpointUpdateRequest(**data)


def test_email_verification_request():
    data = dict(
        recipient='mail@cuenca.com',
        type='email_verification',
        platform_id='PL01',
    )
    with pytest.raises(ValidationError):
        VerificationRequest(**data)
    data['type'] = 'email'
    assert VerificationRequest(**data)


def test_phone_verification_request():
    data = dict(
        recipient='+525555555555',
        type='phone_verification',
        platform_id='PL01',
    )
    with pytest.raises(ValidationError):
        VerificationRequest(**data)
    data['type'] = 'phone'
    assert VerificationRequest(**data)


def test_verification_attempt_request():
    assert VerificationAttemptRequest(**dict(code='111111'))


def test_limited_wallet_request():
    curp = 'TAXM840916HNEMXT02'
    rfc = 'TAXM840916123'
    # Not valid format
    with pytest.raises(ValidationError):
        LimitedWalletRequest(allowed_curp='123', allowed_rfc='123')

    assert LimitedWalletRequest(allowed_curp=curp, allowed_rfc=rfc)


def test_identity_update_request():
    rfc_file = b'this is an example'
    user_id = 'US01'
    extension = 'pdf'

    assert IdentityUpdateRequest(
        user_id=user_id, rfc_file=rfc_file, extension=extension
    )


def test_get_state_name():
    assert get_state_name(State.VZ) == 'Veracruz'


def test_bank_account_validation_clabe_request():
    assert BankAccountValidationRequest(account_number='646180157098510917')



import json
from typing import TYPE_CHECKING, Optional, Type

from pydantic import ConstrainedFloat, ConstrainedInt, ConstrainedStr

from ..validators import sanitize_dict, sanitize_item, validate_digits
from .enums import State

if TYPE_CHECKING:
    from pydantic.typing import CallableGenerator


class SantizedDict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        sanitize_dict(self)


class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        return sanitize_item(o, default=super().default)


class StrictPositiveInt(ConstrainedInt):
    """
    - strict: ensures a float isn't passed in by accident
    - ge (greater than or equal): ensures the value is above 0
    """

    strict = True
    ge = 0
    le = 21_474_836_47  # Max value in DB

    ...


class StrictPositiveFloat(ConstrainedFloat):
    """
    - strict: ensures an integer isn't passed in by accident
    - ge (greater than or equal): ensures the value is above 0
    """

    strict = True
    ge = 0

    ...


def digits(
    min_length: Optional[int] = None, max_length: Optional[int] = None
) -> Type[str]:
    namespace = dict(min_length=min_length, max_length=max_length)
    return type('DigitsValue', (Digits,), namespace)


class Digits(ConstrainedStr):
    @classmethod
    def __get_validators__(cls) -> 'CallableGenerator':
        yield from ConstrainedStr.__get_validators__()
        yield validate_digits


names_state = {
    State.NE: 'Nacido en el Extranjero',
    State.AS: 'Aguascalientes',
    State.BC: 'Baja California',
    State.BS: 'Baja California Sur',
    State.CC: 'Campeche',
    State.CS: 'Chiapas',
    State.CH: 'Chihuahua',
    State.CL: 'Coahuila',
    State.CM: 'Colima',
    State.DF: 'Ciudad de México',
    State.DG: 'Durango',
    State.GT: 'Guanajuato',
    State.GR: 'Guerrero',
    State.HG: 'Hidalgo',
    State.JC: 'Jalisco',
    State.MC: 'México',
    State.MN: 'Michoacan',
    State.MS: 'Morelos',
    State.NT: 'Nayarit',
    State.NL: 'Nuevo León',
    State.OC: 'Oaxaca',
    State.PL: 'Puebla',
    State.QT: 'Querétaro',
    State.QR: 'Quintana Roo',
    State.SP: 'San Luis Potosí',
    State.SL: 'Sinaloa',
    State.SR: 'Sonora',
    State.TC: 'Tabasco',
    State.TL: 'Tlaxcala',
    State.TS: 'Tamaulipas',
    State.VZ: 'Veracruz',
    State.YN: 'Yucatán',
    State.ZS: 'Zacatecas',
}


def get_state_name(state: State):
    return names_state[state]



__all__ = [
    'AccountQuery',
    'Address',
    'ApiKeyQuery',
    'ApiKeyUpdateRequest',
    'AuthorizerTransaction',
    'BalanceEntryQuery',
    'BankAccountValidationQuery',
    'BankAccountValidationRequest',
    'BankAccountStatus',
    'BatchFileMetadata',
    'Beneficiary',
    'BillPaymentQuery',
    'CardErrorType',
    'CardFundingType',
    'CardholderVerificationMethod',
    'CardIssuer',
    'CardNetwork',
    'CardQuery',
    'CardStatus',
    'CardTransactionQuery',
    'CardTransactionType',
    'CardType',
    'Country',
    'CurpField',
    'CurpValidationRequest',
    'CommissionType',
    'DepositNetwork',
    'DepositQuery',
    'EcommerceIndicator',
    'EndpointRequest',
    'EndpointUpdateRequest',
    'EntryType',
    'EventQuery',
    'EventType',
    'FileExtension',
    'FileFormat',
    'FileQuery',
    'FileBatchUploadRequest',
    'FileRequest',
    'FileUploadRequest',
    'Gender',
    'IdentityUpdateRequest',
    'IssuerNetwork',
    'IdentityQuery',
    'JSONEncoder',
    'KYCFile',
    'KYCFileType',
    'KYCValidationRequest',
    'KYCVerificationUpdateRequest',
    'Language',
    'LimitedWalletRequest',
    'PageSize',
    'PaymentCardNumber',
    'PhoneNumber',
    'PlatformRequest',
    'PlatformType',
    'PosCapability',
    'QueryParams',
    'Rfc',
    'SantizedDict',
    'SavingCategory',
    'SavingRequest',
    'SavingUpdateRequest',
    'ServiceProviderCategory',
    'SessionQuery',
    'SessionRequest',
    'SessionType',
    'State',
    'StatementQuery',
    'StrictPaymentCardNumber',
    'StrictPositiveInt',
    'StrictPositiveFloat',
    'StrictTransferRequest',
    'TermsOfService',
    'TOSAgreement',
    'TOSRequest',
    'TrackDataMethod',
    'TransactionQuery',
    'TransactionStatus',
    'TransferNetwork',
    'TransferQuery',
    'TransferRequest',
    'UserCardNotification',
    'UserCredentialRequest',
    'UserCredentialUpdateRequest',
    'UserLoginRequest',
    'UserQuery',
    'UserUpdateRequest',
    'UserRequest',
    'UserStatus',
    'VerificationAttemptRequest',
    'VerificationErrors',
    'VerificationRequest',
    'VerificationStatus',
    'VerificationType',
    'WalletTransactionRequest',
    'WalletTransactionType',
    'WalletQuery',
    'WalletTransactionQuery',
    'WebhookEvent',
    'digits',
    'get_state_name',
]

from .card import PaymentCardNumber, StrictPaymentCardNumber
from .enums import (
    AuthorizerTransaction,
    BankAccountStatus,
    CardErrorType,
    CardFundingType,
    CardholderVerificationMethod,
    CardIssuer,
    CardNetwork,
    CardStatus,
    CardTransactionType,
    CardType,
    CommissionType,
    Country,
    DepositNetwork,
    EcommerceIndicator,
    EntryType,
    EventType,
    FileExtension,
    FileFormat,
    Gender,
    IssuerNetwork,
    KYCFileType,
    Language,
    PlatformType,
    PosCapability,
    SavingCategory,
    ServiceProviderCategory,
    SessionType,
    State,
    TermsOfService,
    TrackDataMethod,
    TransactionStatus,
    TransferNetwork,
    UserCardNotification,
    UserStatus,
    VerificationStatus,
    VerificationType,
    WalletTransactionType,
    WebhookEvent,
)
from .files import BatchFileMetadata
from .general import (
    JSONEncoder,
    SantizedDict,
    StrictPositiveFloat,
    StrictPositiveInt,
    digits,
    get_state_name,
)
from .identities import (
    Address,
    Beneficiary,
    CurpField,
    KYCFile,
    PhoneNumber,
    Rfc,
    TOSAgreement,
    VerificationErrors,
)
from .queries import (
    AccountQuery,
    ApiKeyQuery,
    BalanceEntryQuery,
    BankAccountValidationQuery,
    BillPaymentQuery,
    CardQuery,
    CardTransactionQuery,
    DepositQuery,
    EventQuery,
    FileQuery,
    IdentityQuery,
    PageSize,
    QueryParams,
    SessionQuery,
    StatementQuery,
    TransactionQuery,
    TransferQuery,
    UserQuery,
    WalletQuery,
    WalletTransactionQuery,
)
from .requests import (
    ApiKeyUpdateRequest,
    BankAccountValidationRequest,
    CurpValidationRequest,
    EndpointRequest,
    EndpointUpdateRequest,
    FileBatchUploadRequest,
    FileRequest,
    FileUploadRequest,
    IdentityUpdateRequest,
    KYCValidationRequest,
    KYCVerificationUpdateRequest,
    LimitedWalletRequest,
    PlatformRequest,
    SavingRequest,
    SavingUpdateRequest,
    SessionRequest,
    StrictTransferRequest,
    TOSRequest,
    TransferRequest,
    UserCredentialRequest,
    UserCredentialUpdateRequest,
    UserLoginRequest,
    UserRequest,
    UserUpdateRequest,
    VerificationAttemptRequest,
    VerificationRequest,
    WalletTransactionRequest,
)



import datetime as dt
import re
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, IPvAnyAddress
from pydantic.class_validators import root_validator
from pydantic.types import StrictStr

from .enums import Country, KYCFileType, State, VerificationStatus


class PhoneNumber(StrictStr):
    min_length = 10
    max_length = 15
    regex = re.compile(r'^\+?[0-9]{10,14}$')


class CurpField(StrictStr):
    min_length = 18
    max_length = 18
    regex = re.compile(r'^[A-Z]{4}[0-9]{6}[A-Z]{6}[A-Z|0-9][0-9]$')


class Rfc(StrictStr):
    min_length = 12
    max_length = 13


class Address(BaseModel):
    street: Optional[str] = None
    ext_number: Optional[str] = None
    int_number: Optional[str] = None
    colonia: Optional[str] = None
    postal_code: Optional[str] = None
    state: Optional[State] = None
    country: Optional[Country] = None
    city: Optional[str] = None
    full_name: Optional[str] = None

    class Config:
        schema_extra = {
            "example": {
                "street": "Reforma",
                "ext_number": "265",
                "int_number": "5",
                "colonia": "Cuauhtémoc",
                "postal_code": "06500",
                "state": "DF",
                "country": "MX",
                "city": "Cuauhtémoc",
            }
        }

    @root_validator()
    def full_name_complete(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if values.get('full_name'):
            return values
        if not values.get('street'):
            raise ValueError('required street')
        if not values.get('ext_number'):
            raise ValueError('required ext_number')
        return values


class Beneficiary(BaseModel):
    name: str
    birth_date: dt.date
    phone_number: PhoneNumber
    user_relationship: str
    percentage: int

    class Config:
        schema_extra = {
            "example": {
                "name": "Juan Perez",
                "birth_date": "1907-07-06",
                "phone_number": "+525500998877",
                "user_relationship": "friend",
                "percentage": 100,
            }
        }


class VerificationErrors(BaseModel):
    identifier: str
    error: str
    code: str
    message: Optional[str]

    class Config:
        fields = {
            'identifier': {
                'description': 'Unique identifier for the step validation'
            },
            'error': {
                'description': 'Error throwed on validation,'
                ' can be StepError or SystemError in case of '
                'KYCProvider intermittence'
            },
            'code': {
                'description': 'Specific code of the failure in the step.'
            },
            'message': {'description': 'Error description'},
        }

        schema_extra = {
            "example": {
                "identifier": "age-check",
                "error": 'StepError',
                "code": "underage.noDOB",
                "message": "The date of birth could not be obtained",
            }
        }


class KYCFile(BaseModel):
    type: KYCFileType
    uri_front: str
    uri_back: Optional[str] = None
    is_mx: bool = True
    data: Optional[dict] = None
    status: Optional[VerificationStatus] = None
    errors: Optional[List[VerificationErrors]]
    verification_id: Optional[str]
    attempt: Optional[int]

    class Config:
        fields = {
            'uri_front': {'description': 'API uri to fetch the file'},
            'uri_back': {'description': 'API uri to fetch the file'},
            'status': {
                'description': 'The status of the file depends '
                'on KYCValidation'
            },
            'errors': {
                'description': 'List of document errors found '
                'during kyc validation'
            },
            'attempt': {
                'description': 'The number of kyc_validation '
                'intents for this docuemnt'
            },
            'verification_id': {
                'description': 'The provider identifier of the '
                'validation result'
            },
        }

        schema_extra = {
            "example": {
                "type": "ine",
                "is_mx": True,
                "uri_front": "/files/FILE-01",
                "uri_back": "/files/FILE-02",
                "data": {},
                "status": "created",
                "errors": [],
            }
        }


class TOSAgreement(BaseModel):
    version: str
    ip: IPvAnyAddress
    location: Optional[str]

    class Config:
        schema_extra = {
            "example": {
                "version": "2022-01-01",
                "ip": "192.168.0.1",
                "location": "19.427224, -99.168082",
            }
        }
